library tiger_graph;

import 'package:http/http.dart' as http;
import 'dart:convert';
import 'exceptions.dart';


class FieldValue{
  static final String overwrite = "=";
  static final String increment = "+";
}

class GraphClient implements Editable{

  String _apiKey;
  String _secret;
  String _graph;
  String _server;
  String _urlConstruct = "";
  String _queryConstruct = "";
  final Map<String, String> _authHeaders = {'Authorization': 'Bearer'};
  final List<String> _vertexReqFields = ['vType', 'vId'];

  /*

  Method creates a new Graph Instance with the REST url and OAuth ready to go

  param api: Api Key for Authorization for TigerGraph requests
  param graphName: Name of Graph individual is calling to
  param server: Name of server individual is calling to

   */
  GraphClient(String secret, String graphName, String server){
    this._secret = secret;
    this._graph = graphName;
    this._server = server;
    this._urlConstruct = 'https://$server/graph/$graphName/';
    this._queryConstruct = 'https://$server/query/$graphName/';
  }

  /*

  Methods handles asynchronous connection checking
  Recommended to use rather than the constructor

  param api: Api Key for Authorization for TigerGraph requests
  param graphName: Name of Graph individual is calling to
  param server: Name of server individual is calling to

   */
  static Future<GraphClient> getInstance(String secret, String graphName, String server) async{
    GraphClient g = new GraphClient(secret, graphName, server);
    await g._createApiKey();
    return g;
  }

  /*

  Method creates a temporary auth token

   */
  Future<String> _createApiKey() async {

    String reqUrl = "https://${this._server}/requesttoken?secret=${this._secret}";
    final res = await http.get(reqUrl);
    if(res.statusCode != 200){
      throw Exception("Connection could not be established");
    }
    Map responseBody = json.decode(res.body);

    if(responseBody['error']){
      throw TigerGraphException(responseBody['message']);
    }

    this._apiKey = responseBody['token'];
    this._authHeaders['Authorization'] = 'Bearer ${this._apiKey}';
    return "Success";
  }

  /*

  Method will fetch the TigerGraph query result

  param url: the endpoint with all arguments of the query

   */
  Future<Map> _call(String url) async {
    final res = await http.get(url, headers: this._authHeaders);

    if(res.statusCode != 200){
      throw Exception("Error calling query");
    }

    Map responseBody = json.decode(res.body);

    if(responseBody['error']){
      throw TigerGraphException(responseBody['message']);
    }

    return responseBody;
  }

  /*

  Method will call a TigerGraph query

  param name: Query Name
  param args: the args of the query

   */
  Future<Map> callQuery(String queryName, Map args){

    String reqUrl = this._queryConstruct + "$queryName?";
    List<String> argList = [];
    args.forEach((k,v){
      if(v is List){
        for(int i = 0; i < v.length; i++){
          argList.add("&$k=${v[i]}");
        }
      }
      else if(v is Map){
        String encoding = json.encode(v);
        argList.add("&$k=$encoding");
      }
      else{
        argList.add("&$k=$v");
      }
    });

    reqUrl += argList.join("");

    return this._call(reqUrl);
  }

  /*

  Method grabs vertex and its attributes based on vertex type and id

  param vType: Vertex Type
  param vId: The id of the vertex

   */
  Future<Map> getVertex(Map<String, String> vertex) async {

    assert(vertex != null);
    this._validateInfo(vertex, this._vertexReqFields);
    String reqUrl = this._urlConstruct + "vertices/${vertex['vType']}/${vertex['vId']}";

    final res = await http.get(reqUrl, headers: this._authHeaders);

    if(res.statusCode != 200){
      throw Exception("Connection Error");
    }

    Map responseBody = json.decode(res.body);

    if(responseBody['error']){

      if(responseBody['message'].toString().contains('not a valid vertex id')){
        throw TigerGraphException("Vertex id ${vertex['vId']} does not exist for Vertex type ${vertex['vType']}");
      }
      throw TigerGraphException(responseBody['message']);
    }
    return responseBody;
  }

  /*

  Method returns all edges

  param fromVertex: a map containing data about the starting vertex. For example {'vType': Person, 'vId' : 0990}
  optional param edgeType: a string specifying the edge type of the relationship
  optional param toVertex: a map containing data about the ending vertex. For example {'vType': Event, 'vId': 1002}

  */
  Future<Map> getEdge(Map<String, String> fromVertex, {String edgeType, Map<String, String> toVertex}) async {
    bool isToVertexProvided = false;

    //validate both fromVertex and toVertex, if toVertex exists
    this._validateInfo(fromVertex, this._vertexReqFields);
    if(toVertex != null){
      isToVertexProvided = true;
      this._validateInfo(toVertex, this._vertexReqFields);
    }

    String reqUrl = this._urlConstruct  + "edges/${fromVertex['vType']}/${fromVertex['vId']}/";

    if(edgeType != null){
      reqUrl += "$edgeType/";

      if(isToVertexProvided){
        reqUrl += "${toVertex['vType']}/${toVertex['vId']}";
      }
    }
    else{
      if(isToVertexProvided){
        print("Warning: The Destination Vertex cannot be inferred and thus will be ignored...");
        print("You are seeing this because you have passed in toVertex without edgeType");
      }
    }

    final res = await http.get(reqUrl, headers: this._authHeaders);

    if(res.statusCode != 200){
      throw Exception("Call could not be made");
    }

    Map responseBody = json.decode(res.body);

    if(responseBody['error']){

      if(responseBody['message'].toString().contains('not a valid vertex id')){
        throw TigerGraphException("Either or both vertex id's provided are not valid");
      }

      throw TigerGraphException(responseBody['message']);
    }
    return json.decode(res.body);
  }

  /*

  Method will delete a vertex based on the passed in data and return a string confirming it did

  param vType: The vertex type that shall be deleted
  param vId: The id of the vertex that is intended to be deleted

   */
  Future<String> deleteVertex(Map<String, String> vertex) async {
    this._validateInfo(vertex, this._vertexReqFields);

    String reqUrl = this._urlConstruct + "vertices/${vertex['vType']}/${vertex['vId']}";
    final res = await http.delete(reqUrl, headers: this._authHeaders);

    if(res.statusCode != 200){
      throw Exception("Deleting Vertex cannot be done");
    }

    Map responseBody = json.decode(res.body);

    if(responseBody['error']){
      throw TigerGraphException(responseBody['message']);
    }

    return "Deleted";

  }

  /*

  Method deletes edges based on from vertex

  param fromVertex: a map containing data about the starting vertex. For example {'vType': Person, 'vId' : 0990}
  optional param edgeType: a string specifying the edge type of the relationship
  optional param toVertex: a map containing data about the ending vertex. For example {'vType': Event, 'vId': 1002}

  */
  Future<String> deleteEdge(Map<String, String> fromVertex, {String edgeType, Map<String, String> toVertex, String override}) async {
    bool isToVertexProvided = false;

    //validate both fromVertex and toVertex, if toVertex exists
    this._validateInfo(fromVertex, this._vertexReqFields);
    if(toVertex != null){
      isToVertexProvided = true;
      this._validateInfo(toVertex, this._vertexReqFields);
    }

    String reqUrl = this._urlConstruct  + "edges/${fromVertex['vType']}/${fromVertex['vId']}/";

    if(edgeType != null){
      reqUrl += "$edgeType/";

      if(isToVertexProvided){
        reqUrl += "${toVertex['vType']}/${toVertex['vId']}";
      }
    }
    else{

      if(override == null){
        print("WARNING: EVERY SINGLE EDGE WILL DELETED! If you wish to delete still, pass 'DELETE' for Override");
        return "Didn't delete";
      }

    }

    final res = await http.delete(reqUrl, headers: this._authHeaders);

    if(res.statusCode != 200){
      throw Exception("Deleting and edge cannot happen");
    }

    Map responseBody = json.decode(res.body);

    if(responseBody['error']){
      throw TigerGraphException(responseBody['message']);
    }
    return "Deleted";

  }

  /*

  Method will upsert a vertex based on provided data.
  Recommended if only a couple vertices need to be upserted

  Param vertex: the data of the vertex that shall be upserted

   */
  Future<String> upsertVertex(Map vertex) async {
    final batch = new _WriteBatch(this._urlConstruct, this._apiKey);
    batch.upsertVertex(vertex);
    return await batch.commit();
  }

  /*

  Method will upsert an edge based on provided data.
  Recommended if only a couple edges need to be upserted

  param fromVertex: a map containing data of a vertex where the edge relationship shall being from
  param toVertex:  a map containing data of a vertex where the edge relationship will end
  param relationship: a string denoting the edge relationship between 2 vertices
  param attrs: a map containing the edge attributes

   */
  Future<String> upsertEdge(Map fromVertex, Map toVertex, String relationship, Map attrs) async {
    final batch = new _WriteBatch(this._urlConstruct, this._apiKey);
    batch.upsertEdge(fromVertex, toVertex, relationship, attrs);
    return await batch.commit();
  }

  /*

  Method validates the fields in a provided map

  param requiredFields: a list of the required fields
  param dict: the map that is supposed to provide data

   */
  void _validateInfo(Map dict, List<String> requiredFields){
    for(int i = 0; i < requiredFields.length; i++){
      if(!dict.containsKey(requiredFields[i])){
        throw Exception("${requiredFields[i]} is not provided");
      }
    }
  }

  /*

  Method will return a new WriteBatch which is capable of taking many upsertions in one request

   */
  _WriteBatch createWriteBatch(){
    return _WriteBatch(this._urlConstruct, this._apiKey);
  }

}

//an interface to upsert data in the graph
class Editable{

  upsertVertex(Map e){
    return null;
  }

  upsertEdge(Map v, Map e, String t, Map a){
    return null;
  }
}

//a class to put upsert operations in one request
class _WriteBatch implements Editable{

  String _urlConstruct;
  final Map<String, String> _authHeaders = {'Authorization': 'Bearer'};
  Map _editHistory = {'vertices': {}, 'edges':{}};

  /*

  Method will initialize a new WriteBatch which will allow multiple edits and
  upserts to go through one request

  param url: The url to which the post request will happen to
  param apiKey: The api authorization key

   */
  _WriteBatch(String url, String apiKey){
    this._urlConstruct = url;
    this._authHeaders['Authorization'] += " $apiKey";
  }

  /*

  Method will append the new upsertion to the edit history for a vertex

  param vertexMap: a map containing the metadata as well as the data of the vertex

   */
  //TODO: Implement a way for incrementing and decrementing values
  void upsertVertex(Map vertexMap){

    List<String> vertexFields  = ['vType', 'vId', 'attrs'];

    for(String s in vertexFields){
      if(!vertexMap.containsKey(s)){
        throw Exception("Key $s is not provided");
      }
    }

    //iterating through the edit history and updating changes
    Map current = this._editHistory['vertices'];
    for(int i = 0; i < vertexFields.length - 1 ; i++){
      if(!current.containsKey(vertexFields[i])){
        current[vertexMap[vertexFields[i]]] = {};
      }

      current = current[vertexMap[vertexFields[i]]];
    }

    Map attributeMap = vertexMap['attrs'];

    //iterate through all the attributes and place them inside the edit history
    attributeMap.forEach((k,v){
      current[k] = {'value': v, 'op': '='}; //for now overwrite
    });
  }

  /*

  Method will append the new upsertion to the edit history for an edge

  param fromVertex: a map containing data of a vertex where the edge relationship shall being from
  param toVertex:  a map containing data of a vertex where the edge relationship will end
  param relationship: a string denoting the edge relationship between 2 vertices
  param attrs: a map containing the edge attributes

   */
  void upsertEdge(Map fromVertex, Map toVertex, String relationship, Map attrs){
    List<String> vertexFields  = ['vType', 'vId'];

    for(String s in vertexFields){
      if(!fromVertex.containsKey(s)){
        throw Exception("From Vertex does not contain $s");
      }
      if(!toVertex.containsKey(s)){
        throw Exception("To Vertex does not contain $s");
      }
    }

    List<String> hierarchy = [fromVertex['vType'], fromVertex['vId'], relationship, toVertex['vType'], toVertex['vId']];
    Map current = this._editHistory['edges'];
    for(String s in hierarchy){
      if(!current.containsKey(s)){
        current[s] = {};
      }
      current = current[s];
    }

    attrs.forEach((k,v){
      current[k] = {"value": v, "op": "="}; //overwriting edge attributes for now
    });

  }

  /*

  Method will commit all the edits and upserts to the TigerGraph instance

   */
  Future<String> commit() async {
    
    final res = await http.post(this._urlConstruct, headers: this._authHeaders, body: json.encode(this._editHistory));

    if(res.statusCode != 200){
      throw Exception("Data couldn't be upserted");
    }

    Map responseBody = json.decode(res.body);

    if(responseBody['error']){
      throw Exception(responseBody['message']);
    }

    this._editHistory = {'vertices':{}, 'edges':{}};
    return "Upserted";
  }

}
